import re

from docx import Document
from docx.document import Document as DocumentWithTable
from docx.oxml.table import CT_Tbl
from docx.oxml.text.paragraph import CT_P
from docx.table import _Cell, Table
from docx.text.paragraph import Paragraph

from srsparser import configs
from srsparser.nlp import strings_similarity
from srsparser.sections_tree import SectionsTree


class Parser:
    """
    Класс, анализирующий текстовые документы в формате .docx, содержащие ТЗ, и формирующий деревья разделов на основе
    данных шаблонов.
    """

    def __init__(self, sections_tree_template: dict):
        """
        :param sections_tree_template: шаблон дерева разделов, содержащий в себе определённую структуру дерева
            разделов, которая будет наполняться текстовым содержимым, найденным парсером.
        """
        self.sections_tree = SectionsTree(sections_tree_template)

    def parse_docx(self, path: str) -> dict:
        """
        Возвращает заполненную в соответствии с содержимым текстового документа с ТЗ структуру дерева разделов.

        :param path: путь к текстовому документу в формате .docx, содержащему ТЗ.
        """
        document = Document(path)
        return self.get_docx_structure(document)

    def get_docx_structure(self, doc: Document) -> dict:
        """
        Возвращает заполненную структуру дерева разделов текстового документа с ТЗ.
        """
        sections = self.get_sections_first(doc)
        self.fill_sections_tree(sections)

        sections = self.get_sections_second(doc)
        self.fill_sections_tree(sections)

        return self.sections_tree.to_dict()

    def get_sections_first(self, doc: Document) -> dict:
        """
        Возвращает разделы, основываясь на анализе содержимого абзацев:
        потенциальные заголовки разделов могут быть слева от первого вхождения символа ":"
        (прим.: 1.2 Условное обозначение: АИС «Товарищество собственников жилья»).

        :return: словарь пар вида: "заголовок раздела" — "содержимое заголовка раздела (текст под ним)".
        """
        result = {}

        for paragraph in self.iter_paragraphs(doc):
            p_text_split = paragraph.text.split(":", 1)
            if len(p_text_split) <= 1:
                continue

            if self.is_heading_of_sec(p_text_split[0]) and not self.is_table_element(paragraph, doc):
                if p_text_split[0] != "" and p_text_split[1] != "":
                    result[p_text_split[0]] = p_text_split[1]
        return result

    def get_sections_second(self, doc: Document) -> dict:
        """
        Возвращает разделы, основываясь на положении абзацев в текстовом документе с ТЗ.

        :return: словарь пар вида: "заголовок раздела" — "содержимое заголовка раздела (текст под ним)".
        """
        result = {}
        curr_heading_text = ""
        curr_heading_paragraphs = []

        for paragraph in self.iter_paragraphs(doc):
            # таблицы содержат свои заголовки, поэтому их не берём во внимание
            if self.is_heading_of_sec(paragraph.text) and not self.is_table_element(paragraph, doc):
                if curr_heading_text != "":
                    result[curr_heading_text] = "\n".join(curr_heading_paragraphs)

                curr_heading_text = paragraph.text
                curr_heading_paragraphs.clear()
            else:
                curr_heading_paragraphs.append(re.sub(configs.NUMBERING_PATTERN, "", paragraph.text))
        result[curr_heading_text] = "\n".join(curr_heading_paragraphs)
        return result

    def fill_sections_tree(self, headings_with_texts: dict):
        """
        Заполняет структуру дерева разделов, выраженную уровнями вложенности разделов.
        """
        for heading, text in headings_with_texts.items():
            max_ratio = 0
            text_parent = None
            for section in self.sections_tree.get_leaf_sections():
                ratio = strings_similarity(section.name, heading)
                if ratio >= max_ratio:
                    max_ratio = ratio
                    text_parent = section
            if text_parent is not None:
                text_parent.text = text

    def iter_paragraphs(self, parent):
        if isinstance(parent, DocumentWithTable):
            parent_elm = parent.element.body
        elif isinstance(parent, _Cell):
            parent_elm = parent._tc
        else:
            raise TypeError(repr(type(parent)))

        for child in parent_elm.iterchildren():
            if isinstance(child, CT_P):
                yield Paragraph(child, parent)
            elif isinstance(child, CT_Tbl):
                table = Table(child, parent)
                for row in table.table.rows:
                    for cell in row.cells:
                        for child_paragraph in self.iter_paragraphs(cell):
                            yield child_paragraph

    def is_heading_of_sec(self, p_text: str) -> bool:
        """
        Проверяет, является ли абзац заголовком структуры дерева разделов.

        :return: True — да, иначе — False.
        """
        for section in self.sections_tree.get_leaf_sections():
            if strings_similarity(section.name, p_text) >= configs.MIN_SIMILARITY_RATIO:
                return True
        return False

    @staticmethod
    def is_table_element(paragraph: Paragraph, doc: Document) -> bool:
        """
        Проверяет, является ли абзац элементом таблицы.

        :return: True — да, иначе — False.
        """
        for table in doc.tables:
            for row in table.rows:
                for cell in row.cells:
                    if cell.text.strip() == paragraph.text.strip():
                        return True
        return False



from pymorphy2 import MorphAnalyzer
from gensim.utils import simple_preprocess
from gensim.corpora.dictionary import Dictionary
from gensim.models.tfidfmodel import TfidfModel
from numpy import around

from srsparser import configs
from srsparser.sections_tree import SectionsTree

morph = MorphAnalyzer()


def remove_stop_words(words: list) -> list:
    """
    Удаляет стоп-слова из списка слов.

    :param words: список слов.
    :return: список слов без стоп-слов.
    """
    return list(filter(lambda token: token not in configs.STOPWORDS_RU, words))


def lemmatize(words: list) -> list:
    """
    Приводит каждое слово из списка слов к его лемме.

    :param words: список слов.
    :return: список лемм.
    """
    return list(map(lambda token: morph.normal_forms(token)[0], words))


def exclude_all_except(words: list, part_of_speech: str) -> list:
    """
    Исключает из списка слов все части речи, кроме заданной.

    :param words: список слов.
    :param part_of_speech: сокращённое название части речи на английском языке
        (см. Обозначения для граммем в pymorphy2).
    :return: список слов, содержащий слова одной части речи (part_of_speech).
    """
    return list(filter(lambda word: morph.parse(word)[0].tag.POS == part_of_speech, words))


def tokenize(text: str, part_of_speech="") -> list:
    """
    Предобработка текста: приводит текст к списку токенов.

    Если указан параметр part_of_speech, то при предобработке текста будут исключены все части речи,
    кроме part_of_speech.

    :param text: текст.
    :param part_of_speech: сокращённое название части речи на английском языке
        (см. Обозначения для граммем в pymorphy2).
    :return: список токенов.
    """
    tokens = simple_preprocess(text, min_len=2, max_len=50, deacc=True)
    tokens = remove_stop_words(tokens)
    tokens = lemmatize(tokens)
    if part_of_speech != "":
        tokens = exclude_all_except(tokens, part_of_speech)
    return tokens


def get_tf_idf_weights(structures: list, section_name="Техническое задание", part_of_speech="") -> list:
    """
    Возвращает TF-IDF веса слов для содержимого структур: для каждой структуры "вытаскивает" из разделов текстовое
    содержимое, предобрабатывает его, строит модель TF-IDF и возвращает веса.

    :param structures: список структур ТЗ, листовые разделы которых заполнены содержимым соответствующих текстовых
        документов с ТЗ, взятых из коллекции с результатами парсинга.
    :param section_name: название раздела структуры ТЗ, относительно которого будет выделяться содержимое из каждой
        структуры ТЗ. По умолчанию имеет значение "Техническое задание", так как этот раздел является корневым и
        содержит в себе все разделы структуры ТЗ, но если указать другой раздел, то для анализа будет использоваться
        информация, содержащаяся только в этих разделах данных структур (structures).
    :param part_of_speech: сокращённое название части речи на английском языке (см. Обозначения для граммем в
        pymorphy2). По умолчанию является пустой строкой, поэтому при анализе учитываются все части речи содержимого
        структур с ТЗ, но если указать аббревиатуру конкретной части речи, то из содержимого при предобработке текста
        будут исключены все части речи, кроме данной.
    :return: TF-IDF веса слов для данных структур.
    """
    tokenized = []
    for structure in structures:
        # получаем наполнение структуры (совокупность содержимого всех полей text)
        doc_structure = SectionsTree(structure)
        structure_contents = doc_structure.get_content(section_name)

        # преобразование каждого наполнения структуры в список токенов
        tokenized.append(tokenize(structure_contents, part_of_speech))

    # сохранение извлеченных токенов в словарь
    dictionary = Dictionary(tokenized)

    # преобразование словаря в корпус (собрание документов в виде мешка слов)
    bow_corpus = [dictionary.doc2bow(doc, allow_update=True) for doc in tokenized]

    # TF-IDF веса слов для всех документов (слова, которые часто встречаются в документах, будут понижены в весе)
    tf_idf = TfidfModel(bow_corpus, smartirs="ltc")
    result = []
    for structure in tf_idf[bow_corpus]:
        weight_tf_idf = []
        for id, freq in structure:
            weight_tf_idf.append([dictionary[id], around(freq, decimals=3)])
        result.append(weight_tf_idf)
    return result


def strings_similarity(s1: str, s2: str) -> float:
    """
    Возвращает коэффициент сходимости строк `s1` и `s2`.

    :return: 0.0 — полностью раздичны, 1.0 — одинаковы.
    """
    s1_list = tokenize(s1)
    s2_list = tokenize(s2)

    if not s1_list or not s2_list:
        return 0.0

    l1 = []
    l2 = []

    s1_set = {w for w in s1_list}
    s2_set = {w for w in s2_list}

    # формируется словарь
    rvector = s1_set.union(s2_set)
    for w in rvector:
        if w in s1_set:
            l1.append(1)
        else:
            l1.append(0)
        if w in s2_set:
            l2.append(1)
        else:
            l2.append(0)

    c = 0
    for i in range(len(rvector)):
        c += l1[i] * l2[i]
    cosine = c / float((sum(l1) * sum(l2)) ** 0.5)
    return cosine



from nltk import download, corpus

# регулярное выражение для определения элемента нумерации в строке
NUMBERING_PATTERN = "^([а-я0-9][.) ])+"

# минимальный приемлимый коэффициент сходимости двух строк для того, чтобы считать их похожими
MIN_SIMILARITY_RATIO = 0.5

download("stopwords", quiet=True)
STOPWORDS_RU = corpus.stopwords.words("russian")

SPECIAL_CHARS = "!#$%&'()*+,/;<=>?@[\]^_`{|}~—\"\-."
EXCESS_CHARS = f"[A-Za-z0-9{SPECIAL_CHARS}]+"



import argparse as ap
from os.path import basename, splitext

from docx.opc.exceptions import PackageNotFoundError
from pymongo import MongoClient

from srsparser import Parser


def setup_args_parser():
    parser = ap.ArgumentParser(
        description="Пакет для анализа и загрузки текстовых документов в NoSQL СУБД MongoDB"
    )

    parser.add_argument(
        "connection_string",
        type=str,
        help="строка подключения используется для подключения к MongoDB"
    )

    parser.add_argument(
        "doc_path",
        type=str,
        help="путь к текстовому документу в формате .docx, содержащему ТЗ"
    )

    parser.add_argument(
        "-db",
        dest="db",
        type=str,
        default="documentsAnalysis",
        help="название базы данных MongoDB, "
             "которая содержит в себе коллекции с шаблонами структур ТЗ и результатами парсинга"
    )

    parser.add_argument(
        "-tc",
        dest="tmpl_coll",
        type=str,
        default="sectionTreeTemplates",
        help="название коллекции MongoDB, в которой хранятся шаблоны структур ТЗ, "
             "в соответствии с которыми проводится парсинг текстового документа с ТЗ"
    )

    parser.add_argument(
        "-t",
        dest="tmpl",
        type=str,
        default="default",
        help="название шаблона структуры ТЗ, хранящегося в коллекции tmpl_coll, "
             "в соответствии с которым проводится парсинг текстового документа с ТЗ"
    )

    parser.add_argument(
        "-rc",
        dest="results_coll",
        type=str,
        default="requirementsSpecifications",
        help="название коллекции MongoDB, в которой хранятся результаты анализа текстовых документов с ТЗ "
             "(структуры, заполненные содержанием текстовых документов с ТЗ, "
             "в соответствии с шаблоном и самим документом)"
    )

    return parser


def run(connection_string: str, doc_path: str, db_name: str, tmpl_coll_name: str, tmpl_name: str,
        results_coll_name: str):
    client = MongoClient(connection_string)
    db = client[db_name]

    coll = db[tmpl_coll_name]
    tree_template = coll.find_one({"name": tmpl_name})["structure"]

    parser = Parser(tree_template)
    document_structure = parser.parse_docx(doc_path)

    doc_name = splitext(basename(doc_path))[0]

    coll = db[results_coll_name]
    coll.insert_one({"document_name": doc_name, "structure": document_structure})


def main():
    parser = setup_args_parser()
    args = parser.parse_args()
    try:
        run(args.connection_string, args.doc_path, args.db, args.tmpl_coll, args.tmpl, args.results_coll)
    except PackageNotFoundError:
        print(f"файл не найден по пути: {args.doc_path}")
    except Exception:
        print("неизвестная ошибка")


if __name__ == "__main__":
    main()



from setuptools import setup, find_packages

requirements = [
    'anytree>=2.8.0',
    'python-docx>=0.8.11',
    'pymorphy2>=0.9.1',
    'nltk>=3.6.5',
    'pymongo>=4.0.1',
    'dnspython>=2.1.0',
    'gensim>=4.1.2'
]

setup(
    name='srsparser',
    version='1.0.7',
    author='Курмыза Павел',
    author_email='tmrrwnxtsn@gmail.com',
    url='https://github.com/Text-Analysis/srsparser',
    python_requires='>=3.7',
    description='Пакет для анализа и загрузки текстовых документов в NoSQL СУБД MongoDB.',
    packages=find_packages(),
    install_requires=requirements,
    entry_points={
        'console_scripts': [
            'srsparser = srsparser.__main__:main'
        ]
    },
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        'Topic :: Text Processing :: Linguistic',
    ],
)



from anytree import AnyNode, PreOrderIter, LevelOrderGroupIter
from anytree.exporter import DictExporter
from anytree.importer import DictImporter
from anytree.search import find_by_attr

importer = DictImporter()
exporter = DictExporter()


class Section(AnyNode):
    """
    Класс, представляющий раздел текстового документа с ТЗ.
    """

    def __init__(self, name: str, parent=None, children=None, **kwargs):
        super().__init__(parent, children, **kwargs)
        self.name = name
        self.parent = parent
        if children:
            self.children = children


class SectionsTree:
    """
    Класс, представляющий дерево разделов текстового документа с ТЗ.
    """

    def __init__(self, template: dict):
        """
        :param template: шаблон дерева резделов текстового документа с ТЗ. Можно найти в коллекции MongoDB с шаблонами.
        """
        self.root = importer.import_(template)

    def to_dict(self) -> dict:
        # исключаем из результата листовые разделы, для которых при парсинге не заполнились поля text
        for children in LevelOrderGroupIter(self.root):
            for node in children:
                if hasattr(node, "text") and node.text == "":
                    node.parent = None
        return exporter.export(self.root)

    def get_leaf_sections(self, section_name="Техническое задание") -> list:
        """
        Возвращает список листовых элементов дерева разделов относительно корневого раздела. Если явно указан параметр
        `section_name`, то возвращается список листовых элементов относительно одноимённого раздела.
        """
        search_from_node = find_by_attr(self.root, name="name", value=section_name)
        if search_from_node:
            return [node for node in PreOrderIter(search_from_node, filter_=lambda n: hasattr(n, "text"))]
        return []

    def get_content(self, section_name="Техническое задание") -> str:
        """
        Возвращает содержимое дерева разделов (совокупность значений всех полей `text`) относительно корневого раздела.
        Если указан параметр `section_name`, то возвращается содержимое одноимённого раздела,
        а не всего дерева разделов.
        """
        return "".join([node.text for node in self.get_leaf_sections(section_name)])
